/*----------------------------------------------------------------------------
 Pendula - networked harmonograph and interval guessing game
 
 Copyright (c) 2010 Elie El Noune
 All rights reserved.
 http://ccrma.stanford.edu/~noune/
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 U.S.A.
 -----------------------------------------------------------------------------*/


//-----------------------------------------------------------------------------
// name: pendula.cpp
// 
// author: Elie El Noune
// libraries:
//          (OscPack) Ross Becina
//			(RtAudio) Gary Scavone
// Includes code from: Calories, oscpack examples
// Project for Stanford University | Music 256a: Music, Computing, and Design
//     http://ccrma.stanford.edu/courses/256a/
//-----------------------------------------------------------------------------

#include "RtAudio.h"
#include <sys/time.h>
#include <iostream>
#include <assert.h>
#include <stdio.h>
#include <cstdlib>
#include <string>
#include <string.h>
#include <stdlib.h>
#include "KarplusSynth.h"
#include "defs.h"
#include "RtMidi.h"
#include "MidiEvent.h"
#include "Interval.h"


// threads
#include "util_thread.h"

// STL

#include <vector>

// OpenGL
#if defined(__PLATFORM_MACOSX__)
#include <GLUT/glut.h>
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#else
#include <GL/glut.h>
#include <GL/gl.h>
#include <GL/glu.h>
#endif

// width and height of the window
GLsizei g_width = 800;
GLsizei g_height = 600;

// oscpack
#include "osc/OscReceivedElements.h"
#include "osc/OscPacketListener.h"
#include "osc/OscOutboundPacketStream.h"
#include "ip/UdpSocket.h"


using namespace std;


//globals

//bools

bool g_localPlay = true;
bool g_peerPlay = true;
bool g_keyboard = false;
bool g_server = true;
bool g_connected = false;
bool g_singlePlayer = true;
bool g_processedHome = false;
bool g_processedGuest = false;
bool g_morphing = true;
bool g_pause = false;

//timestamps
int localMidiTime;
int peerMidiTime ;

//osc
UdpTransmitSocket * g_transmit = NULL;
osc::OutboundPacketStream * g_packet = NULL;
const long SEND_SIZE = 4096;
char g_pbuffer[SEND_SIZE];

string g_hostname = "localhost";
//string g_hostname = "171.64.197.164";
long g_osc_port = 8888;
long g_port = 8888;

//midi
MidiEvent * ping = NULL;
MidiEvent * pong = NULL;
KarplusSynth * g_karplusSynth = NULL;
queue<MidiEvent *> localMidiQueue; 
queue<MidiEvent *> peerMidiQueue; 
queue<MidiEvent *> intervalMidiQueue;
interval midiInterval;
interval targetInterval;

//synchronization
XMutex localQ;
XMutex peerQ;
XMutex packet;
XMutex val;

vector<KarplusSynth *> dacs;
vector<SAMPLE *> synthBuffers;
int voices = 16;
//choose voice for new midi event
int current_voice = 0;


// swarmation


void initialize(){
//create 'voices' instances of karplus synths and corresponding buffers to throw their samples in	
	for (int i = 0 ; i < voices ; i++)
	{
		dacs.push_back(new KarplusSynth());
		synthBuffers.push_back(new SAMPLE(BUFFER_SIZE));
	}
	
}
	

void usage( void ) {
	// Error function in case of incorrect command-line
	// argument specifications.
	std::cout << "\nusage: pendula [-s|-c hostname]";
	std::cout << "\nsingle player: pendula\n";
	std::cout << "intervals are relative to C4\n";
	std::cout << "two players, server: pendula -s <hostname>\n";
	std::cout << "two players, client: pendula -c <hostname>\n";
	std::cout << "where hostname = ip address or hostname of peer\n";
	std::cout << "press q to quit\n\n\n";
}


class Game{
	
public: 
	
	int state;
	int stateDurMs;
	int startTime;
	int homeScore;
	int guestScore;
	int hits;
	int misses; 
	int attempts;
	int round; 
	bool initializeState;
	
	Game(): state(0), stateDurMs(10000), startTime(0), initializeState(0), round(0), guestScore(0), homeScore(0), attempts(0), hits(0), misses(0) {};
	
	void setState(int s, int duration ) { state = s; stateDurMs = duration; startTime = glutGet(GLUT_ELAPSED_TIME); }
	
	void startTimer() {
		startTime = glutGet(GLUT_ELAPSED_TIME);
	}
	
	bool timesUp() {
		
		return ( (glutGet(GLUT_ELAPSED_TIME) - startTime) >= stateDurMs);
	}
};

Game swarmation;

void midicallback( double deltatime, std::vector< unsigned char > *message, void *userData )
{
	ping = new MidiEvent(deltatime, message);
	
	localQ.acquire();
	localMidiQueue.push(ping);
	g_processedHome = false;
	localQ.release();
	
	// start bundle
	packet.acquire();
    g_packet->Clear();
    *g_packet << osc::BeginBundleImmediate << osc::BeginMessage( "/Pendula/midi" )
	<< ping->getStamp() << ping->getStatus() << ping->getNote() << ping->getVelocity() << osc::EndMessage
	<< osc::EndBundle;
    g_transmit->Send( g_packet->Data(), g_packet->Size() );
	packet.release();
}

 
void keypresscallback( int note )
{
	double stamp = 0;
	//create a note on event with max velocity
	ping = new MidiEvent(stamp, 144, note, 127);
	
	localQ.acquire();
	localMidiQueue.push(ping);
	g_processedHome = false;
	localMidiTime = glutGet(GLUT_ELAPSED_TIME);
	localQ.release();
	
	packet.acquire();
	// start bundle
    g_packet->Clear();
    *g_packet << osc::BeginBundleImmediate << osc::BeginMessage( "/Pendula/midi" )
	<< ping->getStamp() << ping->getStatus() << ping->getNote() << ping->getVelocity() << osc::EndMessage
	<< osc::EndBundle;
    g_transmit->Send( g_packet->Data(), g_packet->Size() );
	packet.release();
	
}

// audio callback
int callme( void *outputBuffer, void *inputBuffer, unsigned int nBufferFrames,
		   double streamTime, RtAudioStreamStatus status, void *userData )

{	
	int homeNote, guestNote, intervalNote = 0;
		
	SAMPLE * output = (SAMPLE *) outputBuffer;
	//clean the previous buffer
	for (int i = 0; i < nBufferFrames ; i++)
	{
		output[i] = 0;	
	}
	
	
	MidiEvent * home = NULL;
	MidiEvent * guest = NULL;
	MidiEvent * val = NULL;
	
	localQ.acquire();
	peerQ.acquire();
	
	while(!localMidiQueue.empty() || !peerMidiQueue.empty() || !intervalMidiQueue.empty()){

		if (!localMidiQueue.empty()) {
		//home midi event!
		home = localMidiQueue.front();
		//check if note on event
		if (home->getStatus() == 144 && home->getVelocity() != 0 && g_localPlay && g_processedHome)
		{
			homeNote = home->getNote(); 
			//assign event to next dac; (circular array of size voices)
			current_voice %= voices;
			dacs[current_voice]->noteOn(1.0, home->mtof());
			current_voice++;
		}
		
		localMidiQueue.pop();
		}
		
		if (!peerMidiQueue.empty()) {
		//guest midi queue
		guest = peerMidiQueue.front();

		if (guest->getStatus() == 144 && guest->getVelocity() != 0 && g_peerPlay && g_processedGuest)
		{
			guestNote = guest->getNote();
			//assign event to next dac; (circular array of size voices)
			current_voice %= voices;
			dacs[current_voice]->noteOn(1.0, guest->mtof());
			current_voice++;
		}
		
		peerMidiQueue.pop();
		}
		
		if (!intervalMidiQueue.empty()) {
			//home midi event!
			val = intervalMidiQueue.front();
			//check if note on event
			if (val->getStatus() == 144 && val->getVelocity() != 0)
			{
				intervalNote = val->getNote();
				//assign event to next dac; (circular array of size voices)
				current_voice %= voices;
				dacs[current_voice]->noteOn(1.0, val->mtof());
				current_voice++;
			}
			
			intervalMidiQueue.pop();
		}
		
		
	}
	
	
	localQ.release();
	peerQ.release();
	//fill respective buffers
	for (int i=0; i<voices; i++)
	{
		dacs[i]->fillBuffer(outputBuffer, nBufferFrames);
		
	}
	
	    return 0;
}

void sayHello( int count){
	
	for (int i = 0 ; i < count ; i++ )
	{
	packet.acquire();
	// start bundle
	g_packet->Clear();
	*g_packet << osc::BeginBundleImmediate << osc::BeginMessage( "/Pendula/connect" )
	<< 1 << 1 << 1 << 1 << osc::EndMessage
	<< osc::EndBundle;
	g_transmit->Send( g_packet->Data(), g_packet->Size() );
	packet.release();
	}
	
}


//-----------------------------------------------------------------------------
// name: CaloriesPacketListener
// desc: ...
//-----------------------------------------------------------------------------
class CaloriesPacketListener : public osc::OscPacketListener
{
protected:
    virtual void ProcessMessage( const osc::ReceivedMessage & m, 
								const IpEndpointName & remoteEndpoint )
    {
        try{
			static int count = 0;
			
            if( !strcmp( m.AddressPattern(), "/Pendula/midi" ) )
            {
               
				osc::ReceivedMessageArgumentStream args = m.ArgumentStream();
                double a1;
				osc::int32 a2, a3, a4;
                //int a3;
                //int a4;
                args >> a1 >> a2 >> a3 >> a4 >> osc::EndMessage;
				
				//std::cout << "received '/Pendula/midi' message with arguments: "
				//<< a1 << " " << a2 << " " << a3 << " " << a4 << "\n";
				//g_updates++;
				pong = new MidiEvent(a1, a2, a3, a4);
				
				peerQ.acquire();
				peerMidiQueue.push(pong);
				g_processedGuest = false;
				peerMidiTime = glutGet(GLUT_ELAPSED_TIME);
				//cout<<peerMidiTime<<endl;
				peerQ.release();
            }
            else if( !strcmp( m.AddressPattern(), "/Pendula/connect" ) )
            {
				osc::ReceivedMessageArgumentStream args = m.ArgumentStream();
				osc::int32 a1, a2, a3, a4;
			    args>> a1 >> a2 >> a3 >> a4 >> osc::EndMessage;
				
				if (count++ < 1)
				
				{
					cout<<"received connection form pier"<<endl;
					
				}
				if (count++ < 10)
				sayHello(1);
				g_connected = true;
				
				
            }
			
			else if( !strcmp( m.AddressPattern(), "/Pendula/target" ) )
            {
				
				osc::ReceivedMessageArgumentStream args = m.ArgumentStream();
				osc::int32 value;
				//val.acquire();
				args >> value >> osc::EndMessage;
				targetInterval.setInterval(value);
				//val.release();
				//cout<<"received interval"<<targetInterval.getInterval()<<endl;
            }
			
			else if( !strcmp( m.AddressPattern(), "/Pendula/game" ) )
            {
				// this is saying other player has won
				osc::ReceivedMessageArgumentStream args = m.ArgumentStream();
				osc::int32 value, score;
				args >> value >> score >> osc::EndMessage;
				// set Game state to lost!
				peerQ.acquire();
				if (!g_singlePlayer && swarmation.state != 3)
					swarmation.setState(3, 3000);
				// enter state and flush timer
				swarmation.initializeState = false;
				swarmation.guestScore = score;
				peerQ.release();
				
            }
			
                    }catch( osc::Exception& e ){
            std::cerr << "error while parsing message: "
			<< m.AddressPattern() << ": " << e.what() << "\n";
        }
    }
};




//-----------------------------------------------------------------------------
// name: cb_osc_listener()
// desc: ...
//-----------------------------------------------------------------------------
void * cb_osc_listener( void * )
{
    // instantiate listener
    CaloriesPacketListener listener;
    UdpListeningReceiveSocket s(
								IpEndpointName( IpEndpointName::ANY_ADDRESS, g_osc_port ),
								&listener );
	
    // print
    cerr << "OSC listener started on port: " << g_osc_port << "..." << endl;
	
    // go!
    s.RunUntilSigInt();
	
    return NULL;
}

//-----------------------------------------------------------------------------
// GL function prototypes
//-----------------------------------------------------------------------------
void idleFunc( );
void displayFunc( );
void reshapeFunc( GLsizei width, GLsizei height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );

double GetCurrentTimeMs() 
{
	struct timeval t;
	
	gettimeofday( &t, NULL );
	
	return ((double)t.tv_sec ) + ((double)t.tv_usec*0.000001);
//return (((double)t.tv_usec));
}

// entry point
int main( int argc, char ** argv )
{

	usage();
	// parse ip address if provided
	// two player mode 
	if (argc == 3 ) {
		
		if (!strcmp(argv[1], "-s"))
		{
			g_singlePlayer = false;
			g_server = true;
			g_hostname = argv[2];
		}
		else if (!strcmp(argv[1], "-c"))
		{
			g_singlePlayer = false;
			g_server = false;
		g_hostname = argv[2];
		}
		else exit(0);

	}
	// single player client
		else if (argc != 1) exit(0);
	// initialize GLUT
    glutInit( &argc, argv );
    // double buffer, use rgb color, enable depth buffer
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    // initialize the window size
    glutInitWindowSize( g_width, g_height );
    // set the window postion
    glutInitWindowPosition( 100, 100 );
    // create the window
    glutCreateWindow( "Pendula" );
    
    // set the idle function - called when idle
    glutIdleFunc( idleFunc );
    // set the display function - called when redrawing
    glutDisplayFunc( displayFunc );
    // set the reshape function - called when client area changes
    glutReshapeFunc( reshapeFunc );
    // set the keyboard function - called on keyboard events
    glutKeyboardFunc( keyboardFunc );
    // set the mouse function - called on mouse stuff
    glutMouseFunc( mouseFunc );
	// ignore key repeats
	glutIgnoreKeyRepeat(1);
	
	//glEnable(GL_FOG);
	//brown
	//glClearColor(0.35f, 0.25f, 0.15f, 1.0f);
	//dark orange
	//glClearColor(0.75f, 0.45f, 0.15f, 1.0f);
	// fushia
	//glClearColor(0.75f, 0.15f, 0.55f, 1.0f);
	// purple
	//glClearColor(0.45f, 0.15f, 0.55f, 1.0f);
	// deeper purple
	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
	
	g_karplusSynth = new KarplusSynth();
	initialize();
	
	//osc/network
    cout<<"connecting to: "<<g_hostname.c_str()<<":"<< g_port<<endl;
	
    // initialize osc
    g_transmit = new UdpTransmitSocket( IpEndpointName( g_hostname.c_str(), g_port ) );
    g_packet = new osc::OutboundPacketStream( g_pbuffer, SEND_SIZE );
	
	//launching osc listener
	XThread thread;
    thread.start( cb_osc_listener );
	
	//send a flag to other pier
	sayHello(4);
	//RtAudio instantiation and setup
	
	RtAudio dac;
	if ( dac.getDeviceCount() < 1 ) {
		std::cout << "\nNo audio devices found!\n";
		exit( 0 );
	}
	
	RtAudio::StreamParameters parameters;
	parameters.deviceId = dac.getDefaultOutputDevice();
	parameters.nChannels = 1;
	parameters.firstChannel = 0;
	unsigned int sampleRate = SRATE;
	unsigned int bufferFrames = BUFFER_SIZE; 
	double data[2];
	
	try {
		dac.openStream( &parameters, NULL, RTAUDIO_FLOAT64,
					   sampleRate, &bufferFrames, &callme, (void *) &data );
		
		dac.startStream(); 
	}
	catch ( RtError& e ) {
		e.printMessage();
		exit( 0 );
	}    
	
	
    
	//RtMidi instantiation and setup
	
	RtMidiIn *midiin = 0;
	
		// RtMidiIn constructor
	try {
		midiin = new RtMidiIn();
	}
	catch ( RtError &error ) {
		error.printMessage();
		exit( EXIT_FAILURE );
	}
	
	
	//lauch midi now
	
	// Check available ports.
	unsigned int port = 0;
	unsigned int nPorts = midiin->getPortCount();
	if ( nPorts > 0 ) {
		
	try {
		midiin->openPort( port );
	}
	catch ( RtError &error ) {
		error.printMessage();
		goto cleanup;
	}
	}
	
		//always stay in keyboard mode
		g_keyboard = true; 	
	
	
	// Set our callback function.  This should be done immediately after
	// opening the port to avoid having incoming messages written to the
	// queue instead of sent to the callback function.
	midiin->setCallback( &midicallback );
	
	// Don't ignore sysex, timing, or active sensing messages.
	midiin->ignoreTypes( false, false, false );
	
	//std::cout << "\nReading MIDI input ... press <enter> to quit.\n";
	//char input;
	//std::cin.get(input);
	
		
	
	// let GLUT handle the current thread from here
    glutMainLoop();
	
	
		while( true )
    {
        usleep( 10000 );
    }
	
	// if we get here, then stop RtAudio!
    try {
        dac.stopStream();
    } catch( RtError & err ) {
        // do stuff
        err.printMessage();
    }
	
	if ( dac.isStreamOpen() ) 
		dac.closeStream();
	
	// Clean up
cleanup:
	delete midiin;
    
    return 0;
}


//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( GLsizei w, GLsizei h )
{
    // save the new window size
    g_width = w; g_height = h;
    // map the view port to the client area
    glViewport( 0, 0, w, h );
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // create the viewing frustum
    gluPerspective( 45.0, (GLfloat) w / (GLfloat) h, 1.0, 300.0 );
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity( );
    // position the view point
    gluLookAt( 0.0f, 0.0f, -4.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f );
}




//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{	
	//if in not playing freeze keyboard input
	if (key == 'q')
		exit(1);
	
	if (key == 'p')
	{
		g_pause=!g_pause;// change state
		g_keyboard=!g_keyboard;
		if (g_pause)
			glutIdleFunc(NULL);  // turn off animation
		else
			glutIdleFunc(idleFunc);  // turn on animation
		return;		
	}
	if (swarmation.state != 1) {
	
		glutPostRedisplay();
		return; 
	}
    switch( key )
    {
        case 'Q':
        case 'q':
            exit(1);
            break;
			
					
		case 'z':
			if (g_keyboard)
			keypresscallback(60);
			//keypresscallback(60);
			break;
			
		case 's':
			if (g_keyboard)
				keypresscallback(61);
				//keypresscallback(60);
			break;
		
		case 'x':
			if (g_keyboard)
				
				keypresscallback(62);
			    //keypresscallback(60);
			break;
		
		case 'd':
			if (g_keyboard)
				keypresscallback(63);
				//keypresscallback(60);
			break;
		
		case 'c':
			if (g_keyboard)
				keypresscallback(64);
				//keypresscallback(60);
			break;
		
		case 'v':
			if (g_keyboard)
				keypresscallback(65);
				//keypresscallback(60);
			break;
		
		case 'g':
			if (g_keyboard)
				keypresscallback(66);
				//keypresscallback(60);
			break;
		
		case 'b':
			if (g_keyboard)
				keypresscallback(67);
				//keypresscallback(60);
			break;
		
		case 'h':
			if (g_keyboard)
				keypresscallback(68);
				//keypresscallback(60);
			break;
		
		case 'n':
			if (g_keyboard)
				keypresscallback(69);
				//keypresscallback(60);
			break;
		
		case 'j':
			if (g_keyboard)
				keypresscallback(70);
				//keypresscallback(60);
			break;
		
		case 'm':
			if (g_keyboard)
				keypresscallback(71);
				//keypresscallback(60);
			break;
			
		case ',':
			if (g_keyboard)
				keypresscallback(72);
				//keypresscallback(60);
			break;
			
		case 'l':
			if (g_keyboard)
				keypresscallback(73);
				//keypresscallback(60);
			break;
			
		case '.':
			if (g_keyboard)
				keypresscallback(74);
				//keypresscallback(60);
			break;
			
		case ';':
			if (g_keyboard)
				keypresscallback(75);
				//keypresscallback(60);
			break;
			
		case '/':
			if (g_keyboard)
				keypresscallback(76);
				//keypresscallback(60);
			break;
			
		case 'o':
			g_morphing = not(g_morphing);
			break;
			
			
		    }
    
    glutPostRedisplay( );

}




//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y )
{
    if( button == GLUT_LEFT_BUTTON )
    {
        // when left mouse button is down
        if( state == GLUT_DOWN )
        {
        }
        else
        {
        }
    }
    else if ( button == GLUT_RIGHT_BUTTON )
    {
        // when right mouse button down
        if( state == GLUT_DOWN )
        {
        }
        else
        {
        }
    }
    else
    {
    }
    
    glutPostRedisplay( );
}




//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
	
	// render the scene
    glutPostRedisplay( );

}


//-----------------------------------------------------------------------------
// Name: Pendulum
// Desc: Class to create a pendulum object
//-----------------------------------------------------------------------------

class pendulum {
	
public:
	
	double frequency;
	double damping;
	double amplitude;
	double phase;
	
	//default constructor
	pendulum() : frequency(1.00), damping (0.001), amplitude(0.5), phase(0) {};
	//just in case
	pendulum(double f, double d, double a, double p){
		frequency = f; 
		damping = d;
		amplitude = a;
		phase = p;
	}
	double evaluate(GLfloat x, GLfloat freq)
	{
		return amplitude*exp(-damping*x)* sin(2*MY_PIE*freq*x+ phase);
	}
	
	GLfloat evaluate(GLfloat t){
			return amplitude*exp(-damping*t)* sin(2*MY_PIE*frequency*t+ phase);
	}
};

// structure to hold history of harmonographs
struct frequencies{
 double first;
 double second;
 
 frequencies(double x, double y) 
 {first = x; second = y;}
};

//time
//std::cout<<getTimeofDay()<<endl;


vector<frequencies> history;
int history_size = 20;
int history_hop = 1;
static GLfloat col = 0.0f;

//instantiate 4 pendula here: a and c are the lateral pendula, b and d are rotational 
pendulum a, b, c, d, e, f;
//pendula frequencies along x and y axis
double f_x = 2.0;
double f_y = 1.0;
static GLfloat phi_x = 0;
static GLfloat phi_y = 0;
static double framePeriod = 0;
static double currentTimeMs;
static double previousTimeMs = GetCurrentTimeMs();
static GLfloat scaling = 1.0f;
double desiredFP = 1000/50;
static int previousTime, currentTime, timeElapsed;

class harmonogram {

public:
	pendulum a, b, c, d;
	double currentHi;
	double currentLo;
	double targetHi;
	double targetLo;
	double rho;
	GLfloat x, y, z; 
	harmonogram() {
		a = pendulum(1.0, 0.001, 0.5, 0);
		b = pendulum(10.0, 0.001, 0.5, 0);
		c = pendulum(1.0, 0.001, 0.5,  MY_PIE/2);
		d = pendulum(10.0, 0.001, 0.5,  MY_PIE/2);
		x = 0.0f;
		y = 0.0f;
		z = 0.0f;
		currentHi = 1.0f;
		currentLo = 1.0f;
		targetHi = 1.0f;
		targetLo = 1.0f;
		rho = 0.5f;
		
	}
	
	void setFrequencies(GLfloat f1, GLfloat f2, GLfloat f3, GLfloat f4){
				a.frequency = f1;
				b.frequency = f2;
				c.frequency = f3;
				d.frequency = f4;
	}
	
	void setPhases(GLfloat p1, GLfloat p2, GLfloat p3, GLfloat p4){
		a.phase = p1;
		b.phase = p2;
		c.phase = p3;
		d.phase = p4;
	}
	
	void setAmplitude(GLfloat a1, GLfloat a2, GLfloat a3, GLfloat a4){
		a.amplitude = a1;
		b.amplitude = a2;
		c.amplitude = a3;
		d.amplitude = a4;
	}
	
	void setDamping(GLfloat d1, GLfloat d2, GLfloat d3, GLfloat d4){
				a.damping = d1;
				b.damping = d2;
				c.damping = d3;
				d.damping = d4;
	}
	
	void setInt(double higher, double lower){
				a.frequency = 10.0 *(GLfloat)higher+0.0;
				c.frequency = 10.0 *(GLfloat)lower+0.0;
	}
	void draw(GLfloat t){
	
		x = a.evaluate(t)+b.evaluate(t);
		y = c.evaluate(t)+d.evaluate(t);
	}
	
	void drawHarmonogram(interval val, GLfloat t, GLfloat col, double framePeriod)
	{
		if (g_morphing)
		{
		//new interval 
		if (t == 0.0f)
		{
			//cout<<val.getInterval()<<endl;
		    // remember last state
			currentHi = targetHi;
			currentLo = targetLo;
			//cout<<"currentHI"<<currentHi<<endl;
			
			// update new state
			targetHi = val.getHigher();
			//cout<<"targetHi"<<targetHi<<endl;
			targetLo = val.getLower();
			//cout<<"pho"<<rho<<endl;
		}
		
		currentHi  = currentHi + rho* (targetHi-currentHi)*30*framePeriod;;
		currentLo = currentLo+ rho * (targetLo - currentLo)*30*framePeriod;
		
		setInt(currentHi, currentLo);
		} else {
			setInt(val.getHigher(), val.getLower());
		}

		// push the matrix
		glPushMatrix();
		
		//Translate
		glTranslatef( 0.0f, 0.0f, 0.0f );
		
		//glScalef(8.3f, 8.3f, 8.3f);
		
		//Rotate
		
		//glRotatef(rot+= timeElapsed*0.001*rot_inc, 0.0f, 0.0f, 1.0f);
		
		
		glBegin(GL_LINE_STRIP);
		
		for( GLfloat j = 0; j < t; j+= 0.005 )
		{
			// set the color
			glColor3f( sin(0.0125*col+0.001), sin(col*0.025+0.002), sin(0.05*col+0.003) );
			// set local harmonogram to the left
			draw(j);
			//glVertex3f( a.evaluate(j, scaling*a.frequency)+b.evaluate(j, scaling*b.frequency), c.evaluate(j, scaling*c.frequency)+d.evaluate(j, scaling*d.frequency), e.evaluate(j, scaling*e.frequency)+f.evaluate(j, scaling*f.frequency));
			glVertex3f(x, y, z);
			// increment time
			
		}
		
		glEnd();
		//t= 0.0f;
		glPopMatrix();
		
	}	
	
	
};

void printString( string s, GLfloat x, GLfloat y){
	
	glPushMatrix();
	
	char * string = (char *) s.c_str();
	glColor3f(0, 0, 0);
	glRasterPos2f(x, y);
	int len, i;
	len = (int)strlen(string);
	for (i = 0; i < len; i++) {
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, string[i]);
	}
	glPopMatrix();
}	

static harmonogram TheOne;
static harmonogram Target;



//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
	//time processing
	currentTime = glutGet(GLUT_ELAPSED_TIME);
	timeElapsed = currentTime - previousTime;
	previousTime = currentTime;
	//cout<<timeElapsed<<endl;
	
	currentTimeMs = GetCurrentTimeMs();
	framePeriod = (currentTimeMs - previousTimeMs);
	previousTimeMs = currentTimeMs;
	//cout<<framePeriod<<endl;
	//scaling = (GLfloat)framePeriod/desiredFP;
	
	// constant rotation angle 
    static GLfloat rot = -0.5f;
	static string s = ""; 
	static string r = "";
	static string w = "";
	char buffer1[50];
	char buffer2[50];
	char buffer3[50];
	char buffer4[50];
	char buffer5[50];
	char buffer6[50];
	char buffer7[50];
	
	static int homeNote, guestNote = 0;
	//guestNote = 57;
	
    // clear the and depth buffers
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    
    // line width
    glLineWidth( 1.1 );
    // declare time variable 
    static GLfloat t = 0.0f;
	GLfloat decay = 0.0f;
	//GLfloat inc =7.2f/128.0f;
	GLfloat inc =1.0f;
    GLfloat rot_inc = 0.0f;
    GLfloat phase_inc = MY_PIE/240.0f;
    //wuterfall
    GLfloat zinc = 0.8f;
	
	MidiEvent * home = NULL;
	MidiEvent * guest = NULL;
	
	if (g_hostname == "localhost")
	guestNote = 60;
	// mutex lock
	localQ.acquire();
	peerQ.acquire();
	//val.acquire();
	
	// Connection handling
	if (!g_singlePlayer && !g_connected)
	{
		printString("Waiting for connection from pier", 0.5, 0.6);
		
	}
	
	// Game state processing
	
	// if we are initializing 
	 else if ( swarmation.state == 0 )
		
	{
		//only do this once 
		if (!swarmation.initializeState)
		{
			swarmation.round++;	
		swarmation.setState(0, 3000); 
		swarmation.startTimer();
		swarmation.initializeState = true; 
			g_morphing = false;
			t = 0.0f;
		// randomly generate an interval
			
		
			if (g_server){
				
			srand(time(NULL));
			int randInterval = rand()% 11 + 1; 	
			packet.acquire();
			// start bundle
			g_packet->Clear();
			*g_packet << osc::BeginBundleImmediate << osc::BeginMessage( "/Pendula/target" )
			<< randInterval<< osc::EndMessage
			<< osc::EndBundle;
			g_transmit->Send( g_packet->Data(), g_packet->Size() );
			packet.release();			
			
						
			// 2 to 13
			if (targetInterval.compute(1, 1+randInterval))
				s = targetInterval.getInterval();

			
			//play that interval
			//localMidiQueue.push(new MidiEvent(0, 144, 60, 127));
			//localMidiQueue.push(new MidiEvent(0, 144, 60+randInterval, 127));
			
		}
		
		else {
			//cout<<"client"<<endl;
			s = targetInterval.getInterval();
			
		}
		}
		
		// do your thing here
		if (!swarmation.timesUp())
		{
			
			if (!g_server) s = targetInterval.getInterval();
			
			// print round number
			sprintf(buffer2, "ROUND %d", swarmation.round);
			printString( buffer2, 0.15, 1.4);
			
			if (!g_singlePlayer) printString( s, 0.2, 1.1);
			else 
			printString( "New Interval!" , 0.2, 1.1);
			
			// print countdown to the screen
			int countdown =  (3000 - (glutGet(GLUT_ELAPSED_TIME) - swarmation.startTime))/100;
			int seconds = countdown/10;
			int fraction = countdown % 10;
			printString( "STOPWATCH  ", 2.1, -1.3);
			sprintf(buffer4, "%d.%d seconds", seconds, fraction);
			printString( buffer4 , 2.1, -1.5);
			
			// Home Score
			
			if (g_singlePlayer)
			{
				sprintf(buffer1, "HITS: %d", swarmation.homeScore);
				printString( buffer1 , 2.1, 1.4);
				sprintf(buffer3, "MISSES: %d", swarmation.attempts-swarmation.homeScore);
				printString( buffer3 , -1.55, 1.4);
				//compute score as 5* hits- misses;
				sprintf(buffer5, "SCORE: %d", 6*swarmation.homeScore-swarmation.attempts);
				printString( buffer5 , 2.1, 0.0);	
			}
			else 
			{
				sprintf(buffer1, "HITS: %d", swarmation.hits);
				printString( buffer1 , 2.1, 0.0);
				sprintf(buffer3, "MISSES: %d", swarmation.misses);
				printString( buffer3 , -1.55, 0.0);
				
			sprintf(buffer6, "HOME: %d", swarmation.homeScore);
			printString( buffer6 , 2.1, 1.4);
			sprintf(buffer7, "GUEST: %d", swarmation.guestScore);
			printString( buffer7 , -1.55, 1.4);
			}
			// draw target harmonogram
			//GLfloat decay=0.1* exp(-0.4*t);
			
			Target.drawHarmonogram(targetInterval, t, col, framePeriod);
			if (t<1.5)
			t+= framePeriod;
			// playback the interval 
			if (glutGet(GLUT_ELAPSED_TIME) - swarmation.startTime == 700)
			intervalMidiQueue.push(new MidiEvent(0, 144, 60, 127));
			if (glutGet(GLUT_ELAPSED_TIME) - swarmation.startTime == 1500)
			intervalMidiQueue.push(new MidiEvent(0, 144, 60+targetInterval.getNum(), 127)); 
			
		}
		
		else {
		
			swarmation.state = 1; 
			swarmation.initializeState = false;
		}
	}
	
	// playing mode
	else if ( swarmation.state == 1 )
	{
		static GLfloat t2;
		
		if (!swarmation.initializeState)
		{
			swarmation.setState(1, 15000); 
			swarmation.startTimer();
			swarmation.initializeState = true;
			//if (g_singlePlayer) homeNote = 60;
			g_morphing = true;
			t = 0.0f;
			t2 = 0.0f;
			homeNote = 0;
			guestNote = 0;
			s = "";
			
		}
		
		// do you thing here
		if (!swarmation.timesUp())
		{
			// display the target harmonograph on the top right corner 

			if (!localMidiQueue.empty()) {
				//home midi event!
				home = localMidiQueue.front();
				//check if note on event
				if (home->getStatus() == 144 && home->getVelocity() != 0 && !g_processedHome)
				{
					a.frequency = home->mtof()/220.0;
					//reset time
					t= 0.0f;
					swarmation.attempts++;
					homeNote = home->getNote();
					g_processedHome = true;
					if (guestNote != 0) swarmation.misses++;
				}
				//localMidiQueue.pop();
			}
			
			if (!peerMidiQueue.empty() && g_hostname != "localhost") {
				//guest midi queue
				guest = peerMidiQueue.front();
				
				if (guest->getStatus() == 144 && guest->getVelocity() != 0 && g_peerPlay && !g_processedGuest ){
					c.frequency = guest->mtof()/220.0;
				t = 0.0f;
				guestNote = guest->getNote();
					g_processedGuest = true;
				}
				//peerMidiQueue.pop();
			}	
			
			
			if (midiInterval.compute(guestNote, homeNote))
				s = midiInterval.getInterval();
			
			

			if (midiInterval.getNum() == targetInterval.getNum()){
				
				//if (!localMidiQueue.empty()) 
				//localMidiQueue.pop();
				
				//if (!peerMidiQueue.empty())
				//	peerMidiQueue.pop();
				
				// home wins!
				
				if (localMidiTime >= peerMidiTime || g_singlePlayer)
				{
					swarmation.state = 2;
					if (g_singlePlayer)
					swarmation.homeScore++;
					else 
					{
						swarmation.misses--;
						swarmation.hits++;
						swarmation.homeScore = 10 * swarmation.hits - swarmation.misses;
					}
						// broadcast the information to other player
					packet.acquire();
					// start bundle
					g_packet->Clear();
					*g_packet << osc::BeginBundleImmediate << osc::BeginMessage( "/Pendula/game" )
					<< 1 <<swarmation.homeScore<< osc::EndMessage
					<< osc::EndBundle;
					g_transmit->Send( g_packet->Data(), g_packet->Size() );
					packet.release();
				}
				// home loses
					else
						swarmation.state = 3;
				
				swarmation.initializeState = false;
				s = "";
			}
			printString(s, 0.2, -1.3);
			
			// print round number
			sprintf(buffer2, "ROUND %d", swarmation.round);
			printString(buffer2, 0.15, 1.4);
			
			// print countdown to the screen
			int countdown =  (15000 - (glutGet(GLUT_ELAPSED_TIME) - swarmation.startTime))/100;
			int seconds = countdown/10;
			int fraction = countdown % 10;
			printString( "STOPWATCH  ", 2.1, -1.3);
			sprintf(buffer4, "%d.%d seconds", seconds, fraction);
			printString( buffer4 , 2.1, -1.5);
			
			// Home Score
			
			if (g_singlePlayer)
			{
				sprintf(buffer1, "HITS: %d", swarmation.homeScore);
				printString( buffer1 , 2.1, 1.4);
				sprintf(buffer3, "MISSES: %d", swarmation.attempts-swarmation.homeScore);
				printString( buffer3 , -1.55, 1.4);
				// compute score
				sprintf(buffer5, "SCORE: %d", 6*swarmation.homeScore-swarmation.attempts);
				printString( buffer5 , 2.1, 0.0);	
			}
			else 
			{	
			//sprintf(buffer1, "HOME: %d", swarmation.homeScore);
			//printString( buffer1 , 1.9, 1.4);
			//sprintf(buffer3, "GUEST: %d", swarmation.guestScore);
			//printString( buffer3 , -1.5, 1.4);
				sprintf(buffer1, "HITS: %d", swarmation.hits);
				printString( buffer1 , 2.1, 0.0);
				sprintf(buffer3, "MISSES: %d", swarmation.misses);
				printString( buffer3 , -1.55, 0.0);
				
				sprintf(buffer6, "HOME: %d", swarmation.homeScore);
				printString( buffer6 , 2.1, 1.4);
				sprintf(buffer7, "GUEST: %d", swarmation.guestScore);
				printString( buffer7 , -1.55, 1.4);
				
				
			}
			

			//decay=0.1* exp(-0.4*t);
			//t+= decay;
			
			
			TheOne.drawHarmonogram(midiInterval, t, col , framePeriod);
			
			glPushMatrix();
			glTranslatef(-1.75f, -0.7f, 0.0f);
			glScalef(0.42f, 0.42f, 0.45f);
			Target.drawHarmonogram(targetInterval, t2, col, framePeriod );
			glPopMatrix();
			//cout<<t<<endl;
			//t+= timeElapsed * 0.001f;
			if (t < 1.5)
			t+= framePeriod;
			if (glutGet(GLUT_ELAPSED_TIME) - swarmation.startTime < 1500)
			t2+=framePeriod;			
		
		}
		else {
			
			swarmation.state = 3; 
			swarmation.initializeState = false;
			
		}
		
		
	}
		
	// we've got a winner
	else if ( swarmation.state == 2 )
	{
		if (!swarmation.initializeState)
		{
			s = "BINGOOO!!!";
			swarmation.setState(2, 3000); 
			swarmation.startTimer();
			swarmation.initializeState = true; 
			g_morphing = false;
			t= 0.0f;
		}
			// do you thing here
			if (!swarmation.timesUp())
			{
				
				printString( s, 0.2, 1.4);
				
				Target.drawHarmonogram(targetInterval, t, col, framePeriod);
				
				if (t<1.5)
				t+= framePeriod;
				
				if (glutGet(GLUT_ELAPSED_TIME) - swarmation.startTime == 700)
					intervalMidiQueue.push(new MidiEvent(0, 144, 60, 127));
				if (glutGet(GLUT_ELAPSED_TIME) - swarmation.startTime == 1500)
					intervalMidiQueue.push(new MidiEvent(0, 144, 60+targetInterval.getNum(), 127)); 
				if (glutGet(GLUT_ELAPSED_TIME) - swarmation.startTime > 700)
					{
						
						printString( "That was a "+targetInterval.getInterval() , 0.35, 1.1);
						
											
						// Home Score
						
						if (g_singlePlayer)
						{
							sprintf(buffer1, "HITS: %d", swarmation.homeScore);
							printString( buffer1 , 2.1, 1.4);
							sprintf(buffer3, "MISSES: %d", swarmation.attempts-swarmation.homeScore);
							printString( buffer3 , -1.55, 1.4);
							// compute score
							sprintf(buffer5, "SCORE: %d", 6*swarmation.homeScore-swarmation.attempts);
							printString( buffer5 , 2.1, 0.0);	
						}
						else 
						{	
							//sprintf(buffer1, "HOME: %d", swarmation.homeScore);
							//printString( buffer1 , 1.9, 1.4);
							//sprintf(buffer3, "GUEST: %d", swarmation.guestScore);
							//printString( buffer3 , -1.5, 1.4);
							sprintf(buffer1, "HITS: %d", swarmation.hits);
							printString( buffer1 , 2.1, 0.0);
							sprintf(buffer3, "MISSES: %d", swarmation.misses);
							printString( buffer3 , -1.55, 0.0);
							
							sprintf(buffer6, "HOME: %d", swarmation.homeScore);
							printString( buffer6 , 2.1, 1.4);
							sprintf(buffer7, "GUEST: %d", swarmation.guestScore);
							printString( buffer7 , -1.55, 1.4);
							
						}
						

					}
				
			}
			
			else {
				
				swarmation.state = 0; 
				swarmation.initializeState = false;
			}
			
			
	}
	
	// you lost
	else if ( swarmation.state == 3 )
	{
		
		if (!swarmation.initializeState)
		{
			s = "SORRY!! TRY AGAIN!!";
			swarmation.setState(3, 3000); 
			swarmation.startTimer();
			swarmation.initializeState = true; 
			g_morphing = false;
			t = 0.0f;
			
 		}
		// do your thing here
		if (!swarmation.timesUp())
		{
			
			printString( s, 0.3, 1.4);
			Target.drawHarmonogram(targetInterval, t, col, framePeriod);
			if (t<1.5)
			t+= framePeriod;
			if (glutGet(GLUT_ELAPSED_TIME) - swarmation.startTime == 700)
				intervalMidiQueue.push(new MidiEvent(0, 144, 60, 127));
			if (glutGet(GLUT_ELAPSED_TIME) - swarmation.startTime == 1500)
				intervalMidiQueue.push(new MidiEvent(0, 144, 60+targetInterval.getNum(), 127)); 
			if (glutGet(GLUT_ELAPSED_TIME) - swarmation.startTime > 700)
			{
				
				printString( "That was a "+targetInterval.getInterval() , 0.35, 1.1);
				
				
				// Home Score
				
				if (g_singlePlayer)
				{
					sprintf(buffer1, "HITS: %d", swarmation.homeScore);
					printString( buffer1 , 2.1, 1.4);
					sprintf(buffer3, "MISSES: %d", swarmation.attempts-swarmation.homeScore);
					printString( buffer3 , -1.55, 1.4);
					// compute score
					sprintf(buffer5, "SCORE: %d", 6*swarmation.homeScore-swarmation.attempts);
					printString( buffer5 , 2.1, 0.0);	
				}
				else 
				{	
					//sprintf(buffer1, "HOME: %d", swarmation.homeScore);
					//printString( buffer1 , 1.9, 1.4);
					//sprintf(buffer3, "GUEST: %d", swarmation.guestScore);
					//printString( buffer3 , -1.5, 1.4);
					
					sprintf(buffer1, "HITS: %d", swarmation.hits);
					printString( buffer1 , 2.1, 0.0);
					sprintf(buffer3, "MISSES: %d", swarmation.misses);
					printString( buffer3 , -1.55, 0.0);
					
					sprintf(buffer6, "HOME: %d", swarmation.homeScore);
					printString( buffer6 , 2.1, 1.4);
					sprintf(buffer7, "GUEST: %d", swarmation.guestScore);
					printString( buffer7 , -1.55, 1.4);
					
				}
				
			}
			
		}
		
		else {
			
			swarmation.state = 0; 
			swarmation.initializeState = false;
		}
		
	}
	
	else {
		
	

	//cout<<localMidiQueue.size()<<endl;
	if (!localMidiQueue.empty()) {
		//home midi event!
		home = localMidiQueue.front();
		//check if note on event
		if (home->getStatus() == 144 && home->getVelocity() != 0)
		{
			a.frequency = home->mtof()/220.0;
			//reset time
			t= 0.0f;
			//homeNote = home->getNote();
		}
	}		
	
	if (!peerMidiQueue.empty() && g_hostname != "localhost") {
		//guest midi queue
		guest = peerMidiQueue.front();
					
		if (guest->getStatus() == 144 && guest->getVelocity() != 0 && g_peerPlay)
		{
			c.frequency = guest->mtof()/220.0;
			t = 0.0f;
			guestNote = guest->getNote();
		}
	}	
	
	static string s = ""; 
	if (midiInterval.compute(guestNote, homeNote))
			s = midiInterval.getInterval();

	
	glPushMatrix();
		
	char * string = (char *) s.c_str();
	glColor3f(0, 0, 0);
	glRasterPos2f(0.1, 1);
	int len, i;
	len = (int)strlen(string);
	for (i = 0; i < len; i++) {
		glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, string[i]);
	}
	
	glPopMatrix();
	
	GLfloat decay=1.0* exp(-1.0*t);
	t+= decay;
	
	TheOne.drawHarmonogram(midiInterval, t, col, framePeriod);
		
	}
	col+= 15.0f*framePeriod;
	
	//release mutex
	localQ.release();
	peerQ.release();
	//val.release();
    // flush!
    glFlush( );
    // swap the double buffer
    glutSwapBuffers( );
	
}

/*
 
 //GLfloat fogColor[] = {0.15f, 0.75f, 0.15f, 1};
 //glFogfv(GL_FOG_COLOR, fogColor);
 //glFogi(GL_FOG_MODE, GL_LINEAR);
 //glFogf(GL_FOG_START, 8.0f);
 //glFogf(GL_FOG_END, 45.0f);
 
 //save current spectrum to history
 // if history buffer is full, pop oldest element
 if (specHistory.size() == history_size)
 specHistory.pop_back();
 //in all cases, insert the nuest audio buffer at the front
 vector<float *>::iterator front = specHistory.begin(); 
 specHistory.insert(front, temp);
 
 
 //empty the history buffer in hops of history_hop (including current spectrogram)
 for (int i = 0; i<specHistory.size(); i+= history_hop)
 {
 glBegin( GL_LINE_STRIP );
 for( int j = 2; j < g_bufferSize; j++ )
 {
 // set the color
 glColor3f( f2/6.0f, 1.0f-f2/4.0f, 1.0f-sqrt(f2)/3.6f  );
 //draw in 3D: shift x by a constant times the centroid
 glVertex3f(x + 4*c2, specHistory[i][j]- 0.05*pow(i, 1.82), ((float)(-i))*zinc );
 // increment x
 x += inc;
 }
 //reset x to start
 x = -1.8f;
 //cut the line: take off the pen
 glEnd();
 }
 
 glPopMatrix();
 
 // increment color
 //col += .001;
 
 glPushMatrix();
 glScalef(2.3f, 2.3f, 2.3f);
 glTranslatef(0.0f, 0.5f, -5.0);
 glBegin( GL_LINE_STRIP );
 
 // loop through global buffer
 for( int i = 0; i < g_bufferSize; i++ )
 {
 // set the color
 glColor3f( (sin(c)+1)/2, (sin(c*2)+1)/2, (sin(c+.5)+1)/2 );
 // set the next vertex
 glVertex3f( t, 1*windowed [i], -0.0f );
 // increment t
 t += inc;
 }*/

